const crypto = require('crypto');
const { getAESInitValue, MAGIC_SALT, createRandomSalt } = require('./syno_crypto');
const { Transform } = require('stream');
const { Base64Encoder, Base64Decoder } = require('./base64-stream');

const ALGORITHM = 'aes-256-cbc';
const PREFIX_LENGTH = 16;

function encryptStream(stream, password, {format = 'base64'}) {
	const salt = createRandomSalt();
	const {key, iv} = getAESInitValue(salt, password);
	const cipher = crypto.createCipheriv(ALGORITHM, key, iv);
	const prefix = Buffer.alloc(PREFIX_LENGTH);
	prefix.write(MAGIC_SALT);
	prefix.write(salt, 8);
	const outStream = stream.pipe(cipher).pipe(new PrependStream(prefix)).pipe(new Base64Encoder());
	return format === 'json' ? outStream.pipe(new JSONStringStream()) : outStream;
}

function decryptStreamAsync(readable, password) {
	return new Promise((resolve, reject) => {
		const encrypted = readable.pipe(new Base64Decoder());
		let prefix = '';
		function startDecrypt() {
			const salt = prefix.slice(8);
			const {key, iv} = getAESInitValue(salt, password);
			const decipher = crypto.createDecipheriv(ALGORITHM, key, iv);
			resolve(encrypted.pipe(decipher));
		}
		function onEnd() {
			if (prefix.length < PREFIX_LENGTH) {
				reject('data is not enough');
			}
		}
		encrypted.on('readable', function readPrefix() {
			const remind = PREFIX_LENGTH - prefix.length;
			if (remind <= 0) {
				return;
			}
			const data = encrypted.read(remind);
			if (data === null) {
				return;
			}
			prefix += data;
			if (prefix.length === PREFIX_LENGTH) {
				encrypted.removeListener('readable', readPrefix);
				encrypted.removeListener('end', onEnd);
				setImmediate(startDecrypt);
			}
		});
		encrypted.on('end', onEnd);
	});
}

class PrependStream extends Transform {
	constructor(data) {
		super();
		this.data = data;
	}

	_transform(chunk, encoding, callback) {
		if (this.data !== null) {
			this.push(this.data);
			this.data = null;
		}
		this.push(chunk);
		callback();
	}
}

class JSONStringStream extends PrependStream {
	constructor() {
		super('"');
	}

	_flush(callback) {
		this.push('"');
		callback();
	}
}

module.exports = {
	encryptStream,
	decryptStreamAsync,
};
